__all__ = [
    "aslprep_panpipeline",
    "basil_panpipeline",
    "fmriprep_panpipeline",
    "freesurfer_panpipeline",
    "qsiprep_panpipeline",
    "tensor_panpipeline",
    "noddi_panpipeline",
    "panpipeline",
    "dummy_panpipeline",
    "volmeasures_panpipeline",
    "roiextract_panpipeline",
    "textmeasures_panpipeline",
    "collatecsv_panpipeline",
    "collatecsvgroup_panpipeline",
    "lst_panpipeline"
]


from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
from bids import BIDSLayout

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def fmriprep_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    labels_dict = updateParams(labels_dict,"FMRIWORK",cwd + "/fmriwork")
    if not os.path.isdir(cwd + "/fmriwork"):
        os.makedirs(cwd + "/fmriwork")
    labels_dict = updateParams(labels_dict,"FMRIOUTPUT",cwd + "/fmrioutput")
    if not os.path.isdir(cwd + "/fmrioutput"):
        os.makedirs(cwd + "/fmrioutput")
    
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    command_base, container = getContainer(labels_dict,nodename="fmriprep", SPECIFIC="FMRIPREP_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fmriprep version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)


    # handle field maps - use dwirpe as standard if valid, otherwise use megre 
    layout = BIDSLayout(bids_dir)
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')

    fmri_fmap_entity ={}
    fmri_fmap_entity["extension"] = "nii.gz"
    fmri_fmap_entity["datatype"]="fmap"
    fmri_fmap_entity["acquisition"]="fmri"
    fmri_fmap_entity["session"]=participant_session
    fmri_fmap = layout.get(return_type='file', invalid_filters="allow", **fmri_fmap_entity)

    bids_filter_dict={}
    bids_filter_dict["bold"] = {}
    bids_filter_dict["bold"]["session"] =  participant_session
    bids_filter_dict["fmap"] = {}
    bids_filter_dict["fmap"]["session"] =  participant_session

    use_megre_fmap = getParams(labels_dict,'USE_MEGRE_FMAP')
    if use_megre_fmap and participant_label in use_megre_fmap:
        bids_filter_dict["fmap"]["suffix"] =  ["phase1","phase2","magnitude1","magnitude2"]
    elif use_megre_fmap and f"{participant_label}_{participant_session}" in use_megre_fmap:
        bids_filter_dict["fmap"]["suffix"] =  ["phase1","phase2","magnitude1","magnitude2"]
    elif fmri_fmap:
        bids_filter_dict["fmap"]["acquisition"] =  "fmri"
    else:
        bids_filter_dict["fmap"]["suffix"] =  ["phase1","phase2","magnitude1","magnitude2"]

    bids_filter_file = os.path.join(cwd,f"{participant_label}_{participant_session}_bids_filter_file.json")
    export_labels(bids_filter_dict,bids_filter_file)
    IFLOGGER.info(f"Specifying session filter: exporting {bids_filter_dict} to {bids_filter_file}")

    fmriprep_dict={}
    fmriprep_dict = updateParams(fmriprep_dict,"--participant_label","<PARTICIPANT_LABEL>")
    fmriprep_dict = updateParams(fmriprep_dict,"--output-spaces","MNI152NLin6Asym:res-2 MNI152NLin2009cAsym:res-2")
    fmriprep_dict = updateParams(fmriprep_dict,"--skip-bids-validation",IS_PRESENT)
    fmriprep_dict = updateParams(fmriprep_dict,"--bids-filter-file",bids_filter_file)
    fmriprep_dict = updateParams(fmriprep_dict,"--mem_mb","<BIDSAPP_MEMORY>")
    fmriprep_dict = updateParams(fmriprep_dict,"--nthreads","<BIDSAPP_THREADS>")
    fmriprep_dict = updateParams(fmriprep_dict,"--fs-license-file","<FSLICENSE>")
    fmriprep_dict = updateParams(fmriprep_dict,"--omp-nthreads","1")
    fmriprep_dict = updateParams(fmriprep_dict,"-w","/work")

    # Additional params
    FMRIPREP_OVERRIDE_PARAMS = getParams(labels_dict,"FMRIPREP_OVERRIDE_PARAMS")
    if FMRIPREP_OVERRIDE_PARAMS and isinstance(FMRIPREP_OVERRIDE_PARAMS,dict):
        add_labels(FMRIPREP_OVERRIDE_PARAMS,fmriprep_dict)        

    params = ""
    for fmriprep_tag, fmriprep_value in fmriprep_dict.items():
        if "--" in fmriprep_tag and "---" not in fmriprep_tag:
            if fmriprep_value == IS_PRESENT:
                params=params + " " + fmriprep_tag
            elif fmriprep_value == IGNORE:
                IFLOGGER.info(f"Parameter {fmriprep_tag} is being skipped. This has been explicitly required in configuration.")
            else:
                # we dont need = sign for fmriprep just for basi;
                params = params + " " + fmriprep_tag + " " + fmriprep_value

        elif "-" in fmriprep_tag and "--" not in fmriprep_tag:
            params = params + " " + fmriprep_tag + " " + fmriprep_value

        else:
            print(f"fmriprep tag {fmriprep_tag} not valid.")

    command=f"{command_base}"\
            " "+ bids_dir +\
            " /out"\
            " participant"\
            " "+ params 

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    fmri_preprocess_mnilin6 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','func','*space-MNI152NLin6Asym*preproc_bold.nii.gz'))
    fmri_preprocess_mni2009 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','func','*space-MNI152NLin2009cAsym*preproc_bold.nii.gz'))
    mat_t1_mnilin6 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*from-T1w*MNI152NLin6*mode-image_xfm.h5'))
    mat_t1_mni2009 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*from-T1w*MNI152NLin2009*mode-image_xfm.h5'))
    mat_t1_func = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','func','*to-scanner*mode-image_xfm.txt'))
    mat_mnilin6_t1 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*from-MNI152NLin6*mode-image_xfm.h5'))
    mat_mni2009_t1 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*from-MNI152NLin2009*mode-image_xfm.h5'))
    mat_func_t1 = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','func','*from-scanner*mode-image_xfm.txt'))
    confounds = getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','func','*timeseries*.tsv'))
    mnilin6ref =  getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*{}*MNI152NLin6*desc-preproc_T1w.nii.gz'.format(participant_label)))
    mni2009ref =  getGlob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*{}*MNI152NLin2009*desc-preproc_T1w.nii.gz'.format(participant_label)))

    t1ref =  glob.glob(os.path.join(cwd,'fmrioutput','fmriprep','sub-{}'.format(participant_label),'ses-*','anat','*{}*desc-preproc_T1w.nii.gz'.format(participant_label)))
    t1ref = getFirstFromList([ s for s in t1ref if "MNI" not in s])

    output_dir = cwd

    out_files=[]
    out_files.insert(0,fmri_preprocess_mnilin6)
    out_files.insert(1,fmri_preprocess_mni2009)
    out_files.insert(2,mat_t1_mnilin6)
    out_files.insert(3,mat_t1_mni2009)
    out_files.insert(4,mat_t1_func)
    out_files.insert(5,mat_mnilin6_t1)
    out_files.insert(6,mat_mni2009_t1)
    out_files.insert(7,mat_func_t1)
    out_files.insert(8,confounds)
    out_files.insert(9,t1ref)
    out_files.insert(10,mnilin6ref)
    out_files.insert(11,mni2009ref)


    return {
        "fmri_preprocess_mnilin6":fmri_preprocess_mnilin6,
        "fmri_preprocess_mni2009":fmri_preprocess_mni2009,
        "mat_t1_mnilin6":mat_t1_mnilin6,
        "mat_t1_mni2009":mat_t1_mni2009,
        "mat_t1_func":mat_t1_func,
        "mat_mnilin6_t1":mat_mnilin6_t1,
        "mat_mni2009_t1":mat_mni2009_t1,
        "mat_func_t1":mat_func_t1,
        "confounds":confounds,
        "t1ref":t1ref,
        "mnilin6ref":mnilin6ref,
        "mni2009ref":mni2009ref,
        "output_dir":output_dir,
        "out_files":out_files
    }



class fmriprepInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class fmriprepOutputSpec(TraitedSpec):
    fmri_preprocess_mnilin6 = File(desc='Preprocessed DWI MNILin6 space')
    fmri_preprocess_mni2009 = File(desc='Preprocessed DWI MNILin2009 space')
    mat_t1_mnilin6 = File(desc='T1 to MNILin6 transform')
    mat_t1_mni2009 = File(desc='T1 to MNI2009 transform')
    mat_t1_func = File(desc="T1 to Functional transform")
    mat_mnilin6_t1 = File(desc='MNILin6 to T1 transform')
    mat_mni2009_t1 = File(desc='MNI2009 to T1 transform')
    mat_func_t1 = File(desc="Functional to T1 transform")
    confounds = File(desc="fmriprep calculated confounds")
    t1ref = File(desc='T1 reference')
    mnilin6ref = File(desc='MNILin6 reference')
    mni2009ref = File(desc='MNI2009 reference')
    output_dir = traits.String(desc="FMRIPREP output directory")
    out_files = traits.List(desc='list of files')
    
class fmriprep_pan(BaseInterface):
    input_spec = fmriprepInputSpec
    output_spec = fmriprepOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = fmriprep_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='fmriprep_node',bids_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(fmriprep_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
from nipype import logging as nlogging
from scipy.io import savemat 
import gzip
from bids import BIDSLayout
import shutil
import numpy as np

IFLOGGER=nlogging.getLogger('nipype.interface')

def prepare_bidsfile(layout,subject,suffix,datadir):
    mod=layout.get(subject=subject,suffix=suffix, extension='nii.gz')

    if len(mod)>0:
        mod_path=mod[0].path
        mod_file=mod[0].filename
        print("{} image found. Using {}\n\n".format(suffix,mod_file))
        if len(mod) > 1:
            print("Multiple {} images found. Using {}\n\n".format(suffix,mod_file))
            print("{} files found were:".format(mod))
            print('\n'.join([x.path for x in mod]))
 
        try:
            mod_newpath=os.path.join(datadir,mod_file)
            shutil.copyfile(mod_path,mod_newpath)
            print("Copying file from {} to {}".format(mod_path,mod_newpath))

            mod_newpath_nii=mod_newpath.split('.nii.gz')[0] + '.nii'
            decompress(mod_newpath,mod_newpath_nii,"gzip-clean")
            print("Decompressing {} in place to {}".format(mod_newpath,mod_newpath_nii))
            return mod_newpath_nii
        except Exception as e:
            print("Problem copying {} to {} and decompressing to {}".format(mod_path,mod_newpath,mod_newpath_nii))
            print(e)
            return None

    else:
        print("{} image not found".format(mod))
        return None

def create_lpa(matfile,flair,OPT=''):
    matlabbatch = np.zeros((1,),dtype=object) 
    matlabbatch[0]={}
    matlabbatch[0]['spm']={}
    matlabbatch[0]['spm']['tools']={}
    matlabbatch[0]['spm']['tools']['LST']={}
    matlabbatch[0]['spm']['tools']['LST']['lpa']={}

    file2 = np.zeros((1,),dtype=object)
    file2[0]='{},1'.format(flair)
    matlabbatch[0]['spm']['tools']['LST']['lpa']['data_F2']=[file2]

    opt = np.zeros((1,),dtype=object) 
    opt[0]=OPT         
    matlabbatch[0]['spm']['tools']['LST']['lpa']['data_coreg']=[opt]
    matlabbatch[0]['spm']['tools']['LST']['lpa']['html_report']=1 
    savemat(matfile,{'matlabbatch':matlabbatch})
    print("Created {}".format(matfile))


def create_lga(matfile,t1w,flair,INIT=300,MRF=1,MAXITER=50,HTML=1):
    matlabbatch = np.zeros((1,),dtype=object) 
    matlabbatch[0]={}
    matlabbatch[0]['spm']={}
    matlabbatch[0]['spm']['tools']={}
    matlabbatch[0]['spm']['tools']['LST']={}
    matlabbatch[0]['spm']['tools']['LST']['lga']={}

    file1 = np.zeros((1,),dtype=object)
    file1[0]='{},1'.format(t1w)
    matlabbatch[0]['spm']['tools']['LST']['lga']['data_T1']=[file1]

    file2 = np.zeros((1,),dtype=object)
    file2[0]='{},1'.format(flair)
    matlabbatch[0]['spm']['tools']['LST']['lga']['data_F2']=[file2]
                
    matlabbatch[0]['spm']['tools']['LST']['lga']['opts_lga']={}
    matlabbatch[0]['spm']['tools']['LST']['lga']['opts_lga']['initial']=INIT
    matlabbatch[0]['spm']['tools']['LST']['lga']['opts_lga']['mrf']=MRF  
    matlabbatch[0]['spm']['tools']['LST']['lga']['opts_lga']['maxiter']=MAXITER
    matlabbatch[0]['spm']['tools']['LST']['lga']['html_report']=HTML 
    savemat(matfile,{'matlabbatch':matlabbatch})
    print("Created {}".format(matfile))

def decompress(infile, outfile, type):
    if type == 'gzip-clean':
        with open(infile,'rb') as fi:
            indata=fi.read()

        os.remove(infile)
        
        indata_dec=gzip.decompress(indata)
        with open(outfile,'wb') as fo:
            fo.write(indata_dec)

def compress(infile, outfile, type):
    if type == 'gzip-clean':
        with open(infile,'rb') as fi:
            with gzip.open(outfile,'wb') as fo:
                shutil.copyfileobj(fi,fo)
        os.remove(infile)

    elif type == 'gzip':
        with open(infile,'rb') as fi:
            with gzip.open(outfile,'wb') as fo:
                shutil.copyfileobj(fi,fo)


def lst_proc(labels_dict,bids_dir=""):
    
    command_base, container = getContainer(labels_dict,nodename="lst",SPECIFIC="LST_CONTAINER",LOGGER=IFLOGGER)

    cwd=os.getcwd()
    lst_type = getParams(labels_dict,"LST_TYPE")
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    output_dir=os.path.join(cwd,'sub-' + participant_label,lst_type)
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    data_dir=os.path.join(output_dir,'data')
    if not os.path.isdir(data_dir):
        os.makedirs(data_dir)

    labels_dict = updateParams(labels_dict,"CWD",cwd)
    labels_dict = updateParams(labels_dict,"LST_OUTPUT_DIR",output_dir)

    IFLOGGER.info(f"Running the {lst_type} algorithm for {participant_label}")

    layout = BIDSLayout(bids_dir)
    if lst_type == "LGA":
        T1w = prepare_bidsfile(layout,participant_label,'T1w',data_dir)
        FLAIR =  prepare_bidsfile(layout,participant_label,'FLAIR',data_dir)
        if FLAIR is None or T1w is None:
            IFLOGGER.info(f"FLAIR or T1w is missing. Not enough data to run {lst_type}. Returning")
            return
        else:
            matfile=os.path.join(output_dir,participant_label + f'_{lst_type}.mat')
            create_lga(matfile,T1w,FLAIR)

    elif lst_type == 'LPA':
        FLAIR =  prepare_bidsfile(layout,participant_label,'FLAIR',data_dir)
        if FLAIR is None :
            IFLOGGER.info(f"FLAIR is missing. Not enough data to run {lst_type}. Returning")
            return
        else:
            matfile=os.path.join(output_dir,participant_label + f'_{lst_type}.mat')
            create_lpa(matfile,FLAIR)

    params = f"{matfile}"   

    command = f"{command_base} run_jobman"\
        " "+params 

    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    ples_nii = getGlob(os.path.join(data_dir,'ples_*.nii'))
    ples = ples_nii + '.gz'
    compress(ples_nii,ples,"gzip-clean")

    msub_nii = getGlob(os.path.join(data_dir,'*msub*.nii'))
    msub = msub_nii + '.gz'
    compress(msub_nii,msub,"gzip-clean")
    

    out_files=[]
    out_files.insert(0,ples)
    out_files.insert(1,msub)


    return {
        "ples": ples,
        "msub": msub,
        "output_dir":output_dir,
        "out_files":out_files
    }



class lstInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class lstOutputSpec(TraitedSpec):
    ples= File(desc='ples')
    msub = File(desc='msub')
    output_dir = traits.String(desc="lst output directory")
    out_files = traits.List(desc='list of files')
    
class lst_pan(BaseInterface):
    input_spec = lstInputSpec
    output_spec = lstOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = lst_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='lst_node',bids_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(lst_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
import pandas as pd
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def roi_mean_group_proc(labels_dict,file_template,atlas_file,atlas_index):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir=cwd
    # retrieve subject list
    fsldesign_text = getParams(labels_dict,"TEXT_FSL_DESIGN")
    df = pd.read_table(fsldesign_text,sep=",",header=None)


    command_base, container = getContainer(labels_dict,nodename="roi_mean_group", SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    statsimagestring=''
    stats_image=None
    for subindex in range(len(df)):
        participant_label=df[0][subindex].split('sub-')[1]
        labels_dict = updateParams(labels_dict,'PARTICIPANT_LABEL',participant_label)
        subject_file=getGlob(substitute_labels(getParams(labels_dict,"STAT_FILE_TEMPLATE"),labels_dict))
        statsimagestring= statsimagestring + " " + subject_file 

    if not statsimagestring == "":
        stats_image_dir = os.path.join(cwd,'stats_image')
        if not os.path.isdir(stats_image_dir):
            os.makedirs(stats_image_dir)
        stats_image = os.path.join(stats_image_dir,'stats_image_mni.nii.gz')

        params="-t" \
            " "+ stats_image + \
            " " + statsimagestring 

        command=f"{command_base} fslmerge"\
            " "+params

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)


    
    everythingThere=True

    if everythingThere:
        roi_output_dir = os.path.join(cwd,'roi_output_dir')
        if not os.path.isdir(roi_output_dir):
            os.makedirs(roi_output_dir)

        roi_raw_txt = os.path.join(roi_output_dir,'roi_raw.txt')

        params = " -i "+stats_image+ \
            " -o "+roi_raw_txt+\
            " --label="+atlas_file

        command=f"{command_base} fslmeants"\
            " "+params

        evaluated_command=substitute_labels(command,labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)


    with open(atlas_index, 'r') as in_file:
        lines = in_file.readlines()

    table_columns = [x.replace('\n','') for x in lines]
    df2 = pd.read_table(roi_raw_txt,sep=r"\s+",header=None)
    df2.insert(0,"subject_id",df[0].tolist())
    table_columns.insert(0,"subject_id")
    df2.columns = table_columns
    
    roi_csv = os.path.join(roi_output_dir,'{}_{}.csv'.format(get_modality(file_template),os.path.basename(atlas_file).split('.')[0]))
    df2.to_csv(roi_csv,sep=",",header=True, index=False)


    out_files=[]
    out_files.insert(0,roi_csv)
    out_files.insert(1,stats_image)

    return {
        "roi_csv":roi_csv,
        "merged_image":stats_image,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class roi_mean_groupInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    file_template = traits.String("",desc="file template", usedefault=True)
    atlas_file = File(desc='atlas file')
    atlas_index = File(desc='atlas index file')

class roi_mean_groupOutputSpec(TraitedSpec):
    roi_csv = File(desc='CSV file of results')
    merged_image = File(desc='4D stats image')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class roi_mean_group_pan(BaseInterface):
    input_spec = roi_mean_groupInputSpec
    output_spec = roi_mean_groupOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = roi_mean_group_proc(
            self.inputs.labels_dict,
            self.inputs.file_template,
            self.inputs.atlas_file,
            self.inputs.atlas_index,
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="roi_mean_group_node",file_template="",atlas_file="",atlas_index="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(roi_mean_group_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
            
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if file_template is None or file_template == "":
        file_template = getParams(labels_dict,"STAT_FILE_TEMPLATE")
    pan_node.inputs.file_template =  file_template

    if atlas_file is None or atlas_file == "":
        atlas_file = getParams(labels_dict,"ATLAS_FILE")
    pan_node.inputs.atlas_file =  atlas_file

    if atlas_index  is None or atlas_index  == "":
        atlas_index = getParams(labels_dict,"ATLAS_INDEX")
    pan_node.inputs.atlas_index =  atlas_index

    return pan_node





import json
from shutil import copyfile
from pathlib import Path
from functools import partial
import os

def _path_exists(path, parser):
    """Ensure a given path exists."""
    if path is None or not Path(path).exists():
        raise parser.error(f"Path does not exist: <{path}>.")
    return Path(path).expanduser().absolute()

def get_parser():
    from argparse import ArgumentParser
    from argparse import RawTextHelpFormatter

    parser = ArgumentParser(description="Get Templare from template flow.")
    PathExists = partial(_path_exists, parser=parser)
    parser.add_argument("TemplateFlowDir", type=PathExists, help="The directory where templatefiles will be stored")
    parser.add_argument('--template_dict', action='store',type=json.loads,
        help='Template parameters as json dictionary.')

    return parser


def getTemplateRef(TEMPLATEFLOW_HOME,template_space,suffix=None,desc=None,resolution=None,extension=[".nii.gz"]):
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    from templateflow import api as tf

    template_ref=tf.get(template_space,resolution=resolution,desc=desc,suffix=suffix,extension=extension)
    return template_ref


def main():

    opts = get_parser().parse_args()

    template_dict = None
    if opts.template_dict:
        template_dict = opts.template_dict

    template= "MNI152NLin6Asym"
    if "template" in template_dict:
        template = template_dict["template"]

    suffix = None
    if "suffix" in template_dict:
        suffix = template_dict["suffix"]

    desc = None
    if "desc" in template_dict:
        desc = template_dict["desc"]

    resolution = None
    if "resolution" in template_dict:
        resolution = template_dict["resolution"]

    extension = [".nii.gz"]
    if "extension" in template_dict:
        extension = template_dict["extension"]

    templateRef = getTemplateRef(str(opts.TemplateFlowDir),template,suffix=suffix,desc=desc,resolution=resolution,extension=extension)
    templateCurr = os.path.join(os.getcwd(),os.path.basename(templateRef))
    copyfile(templateRef,templateCurr)



# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()



#!/usr/bin/env python
from panpipelines.utils.util_functions import *
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
import json
import pandas as pd
import os
import datetime
import logging
import sys
from panpipelines.version import __version__
from panpipelines import Factory

LOGGER = logger_setup("panpipelines.single_subject", logging.DEBUG)
logger_addstdout(LOGGER, logging.INFO)

panFactory = Factory.getPANFactory()

def runSingleSubject(participant_label, xnat_project, session_label, pipeline, pipeline_class, pipeline_outdir, panpipe_labels,bids_dir,cred_user,cred_password, execution_json,analysis_level="participant",panlabel=None):

    pipeline_start = datetime.datetime.now()
    LOGGER.info("---------------------------------------------------------------------------------")
    LOGGER.info(f"Single Subject Processing for {participant_label} started at: {str(pipeline_start)}")
    LOGGER.info("---------------------------------------------------------------------------------")

    panpipe_labels = updateParams(panpipe_labels,"PARTICIPANT_LABEL",participant_label)
    panpipe_labels = updateParams(panpipe_labels,"PARTICIPANT_XNAT_PROJECT",xnat_project)
    panpipe_labels = updateParams(panpipe_labels,"PARTICIPANT_SESSION",session_label)

    if not panlabel:
        # get parent directory
        panlabel=os.path.basename(os.path.dirname(pipeline_outdir))

    pipeline_outdir=os.path.join(pipeline_outdir,xnat_project)
    if not os.path.exists(pipeline_outdir):
        os.makedirs(pipeline_outdir,exist_ok=True)

    participant_index = getParams(panpipe_labels,"PARTICIPANT_INDEX")
    if participant_index is None:
        participant_index = ""
    else:
        participant_index = "_" + str(participant_index)
        
    if not session_label:
        LOGFILE=os.path.join(pipeline_outdir,f"{participant_label}{participant_index}_{xnat_project}_{pipeline}_{panlabel}.log")
    else:
        LOGFILE=os.path.join(pipeline_outdir,f"{participant_label}{participant_index}_{session_label}_{xnat_project}_{pipeline}_{panlabel}.log")
    logger_addfile(LOGGER, LOGFILE, logging.DEBUG)
    nipype_loggers_setup(logging.INFO,LOGFILE,logging.DEBUG)

    LOGGER.info(f"Pan Processing for single subject: Running {pipeline} pipeline for {participant_label}")
    LOGGER.info(f"start logging to {LOGFILE}")

    
    getSubjectBids(panpipe_labels,bids_dir,participant_label,xnat_project,cred_user,cred_password)

    panProcessor = panFactory.get_processflow(pipeline_class)

    if not session_label:
        pipeline_outdir_subject = os.path.join(pipeline_outdir,"sub-"+participant_label)
    else:
        pipeline_outdir_subject = os.path.join(pipeline_outdir,"sub-"+participant_label,"ses-"+session_label)

    PanProc = panProcessor(panpipe_labels,pipeline_outdir_subject, participant_label, name=pipeline,LOGGER=LOGGER,execution=execution_json,analysis_level=analysis_level,participant_project=xnat_project, participant_session=session_label)
    PanProc.run()

    pipeline_end = datetime.datetime.now()
    LOGGER.info("---------------------------------------------------------------------------------")
    LOGGER.info(f"Single Subject Processing for {participant_label} Completed at: {str(pipeline_end)}")
    LOGGER.info("---------------------------------------------------------------------------------")
    pipeline_duration = pipeline_end - pipeline_start
    LOGGER.info(f"Single Subject Processing for {participant_label} Duration: {str(pipeline_duration)}")
    LOGGER.info("---------------------------------------------------------------------------------")

    LOGGER.debug(f"\nDump of configuration settings for {participant_label} run of {pipeline}")
    LOGGER.debug("---------------------------------------------------------------------------------")
    LOGGER.debug(f"{panpipe_labels!r}")
    LOGGER.debug("\n------  environment dump -----------")
    LOGGER.debug(os.environ)

def parse_params():
    parser = ArgumentParser(description="Pan pipelines")
    PathExists = partial(path_exists, parser=parser)
    parser.add_argument("config_file", type=PathExists, help="Pipeline Configuration File")
    parser.add_argument("pipeline", help="single running pipeline")
    parser.add_argument("credentials", help="credential file")
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    return parser

def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()

    print(f"Running {__file__} v{__version__}")

    panpipe_labels={}
    panpipe_labels = updateParams(panpipe_labels,'PWD',str(os.getcwd()))

    panpipeconfig_file=str(args.config_file)
    panpipeconfig_json=None
    if os.path.exists(panpipeconfig_file):
        with open(panpipeconfig_file, 'r') as infile:
            panpipeconfig_json = json.load(infile)
    panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,panpipe_labels,uselabel=False)

    pipeline=args.pipeline

    pipeline_class = getParams(panpipe_labels,"PIPELINE_CLASS")
    if pipeline_class is None:
        pipeline_class = pipeline

    #use_pipeline_desc = getParams(panpipe_labels,"USE_PIPELINE_DESC")
    #if use_pipeline_desc is None:
    #    use_pipeline_desc = ""

    #pipeline_desc = getParams(panpipe_labels,"PIPELINE_DESC")
    #if pipeline_desc is None or use_pipeline_desc == "N":
    #    pipeline_desc = pipeline 
    #else:
    #    pipeline_desc = "".join([x if x.isalnum() else "_" for x in pipeline_desc]) 

    pipeline_outdir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),pipeline)
    if not os.path.exists(pipeline_outdir):
        os.makedirs(pipeline_outdir,exist_ok=True)

    lock_dir=getParams(panpipe_labels,"LOCK_DIR")
    if lock_dir:
        if not os.path.exists(lock_dir):
            os.makedirs(lock_dir,exist_ok=True)
    else:
        lock_dir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),"datalocks")
        if not os.path.exists(lock_dir):
            os.makedirs(lock_dir,exist_ok=True)
        panpipe_labels = updateParams(panpipe_labels,"LOCK_DIR",lock_dir)

    credentials = os.path.abspath(args.credentials)
    if os.path.exists(credentials):
        with open(credentials, 'r') as infile:
            cred_dict = json.load(infile)
            cred_user = getParams(cred_dict,"user")
            cred_password = getParams(cred_dict,"password")
    else:
        cred_user = "dummy_user"
        cred_password = "dummy_password"

    execution_json=getParams(panpipe_labels,"NIPYPE_CONFIG")
    if execution_json is None:
        execution_json = {} 


    participants_file = getParams(panpipe_labels,"PARTICIPANTS_FILE")
    sessions_file = getParams(panpipe_labels,"SESSIONS_FILE")
    bids_dir = getParams(panpipe_labels,"BIDS_DIR")

    # if participant label not provided then infer as submitted job
    ARRAY_INDEX = getParams(panpipe_labels,"ARRAY_INDEX")
    if ARRAY_INDEX is not None and ARRAY_INDEX in os.environ.keys():
        participant_index = int(os.environ[ARRAY_INDEX])
        panpipe_labels = updateParams(panpipe_labels, "PARTICIPANT_INDEX",str(participant_index))
        if not sessions_file:
            df = pd.read_table(participants_file,sep="\t")
        else:
            df = pd.read_table(sessions_file,sep="\t")
        if participant_index <= len(df):
            participant_label = drop_sub(df['bids_participant_id'].iloc[participant_index - 1])
            xnat_project = df['project'].iloc[participant_index - 1]
            if sessions_file:
                session_label = drop_ses(df['bids_session_id'].iloc[participant_index - 1])
            else:
                session_label=None
            runSingleSubject(participant_label,xnat_project,session_label, pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level="participant")


        else:
            print("Problem: ARRAY_INDEX {} greater than length of participants file {}".format(participant_index,participants_file))
    else:
        print("{} used as ARRAY_INDEX but not defined in os.environs".format(ARRAY_INDEX))


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()


from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def mni3dglm_proc(labels_dict,file_template,mask_template,design_file,contrast_file,ftest_file):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    output_dir=cwd
    # retrieve subject list
    fsldesign_text = getParams(labels_dict,"TEXT_FSL_DESIGN")
    df = pd.read_table(fsldesign_text,sep=",",header=None)

    command_base, container = getContainer(labels_dict,nodename="fslanat", SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    groupmaskdata=None
    for subindex in range(len(df)):
        participant_label=df[0][subindex].split('sub-')[1]
        labels_dict = updateParams(labels_dict,'PARTICIPANT_LABEL',participant_label)
        subject_mask=substitute_labels(getParams(labels_dict,"STAT_MASK_TEMPLATE"),labels_dict)

        img = nib.load(subject_mask)
        img_dtype = img.header.get_data_dtype()
        data=img.get_fdata()
        if subindex == 0:
            groupmaskdata = data.copy()
        else:
            groupmaskdata = groupmaskdata + data

    if groupmaskdata is not None:
        group_mask_dir = os.path.join(cwd,'group_mask')
        if not os.path.isdir(group_mask_dir):
            os.makedirs(group_mask_dir)
        group_mask = os.path.join(group_mask_dir,'group_mask_mni.nii.gz')
        groupmaskdata[groupmaskdata > 0] = 1
        save_image_to_disk(subject_mask,groupmaskdata,group_mask)

        command=f"{command_base} fslreorient2std"\
            " "+group_mask+ " "+group_mask

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        command=f"{command_base} fslmaths"\
            " "+group_mask+ " -bin "+group_mask

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        command=f"{command_base} fslmaths"\
            " "+group_mask+ " -dilM -fillh "+group_mask

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    statsimagestring=''
    stats_image=None
    for subindex in range(len(df)):
        participant_label=df[0][subindex].split('sub-')[1]
        labels_dict = updateParams(labels_dict,'PARTICIPANT_LABEL',participant_label)
        subject_file=getGlob(substitute_labels(getParams(labels_dict,"STAT_FILE_TEMPLATE"),labels_dict))
        statsimagestring= statsimagestring + " " + subject_file 

    if not statsimagestring == "":
        stats_image_dir = os.path.join(cwd,'stats_image')
        if not os.path.isdir(stats_image_dir):
            os.makedirs(stats_image_dir)
        stats_image = os.path.join(stats_image_dir,'stats_image_mni.nii.gz')

        params="-t" \
            " "+ stats_image + \
            " " + statsimagestring 

        command=f"{command_base} fslmerge"\
            " "+params

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    
    # permutation
    everythingThere=True

    if everythingThere:
        stats_output_dir = os.path.join(cwd,'stats_results')
        if not os.path.isdir(stats_output_dir):
            os.makedirs(stats_output_dir)

        PERMS="50"
        params = " -i "+stats_image+ \
            " -o "+stats_output_dir+"/out"\
            " -m "+group_mask+\
            " -d "+design_file+\
            " -t "+contrast_file+\
            " -f "+ftest_file+\
            " -n "+PERMS+\
            " -T"

        command=f"{command_base} randomise_parallel"\
            " "+params

        evaluated_command=substitute_labels(command,labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    out_files=[]
    out_files.insert(0,group_mask)
    out_files.insert(1,stats_image)

    return {
        "group_mask":group_mask,
        "stats_image":stats_image,
        "stats_output_dir":stats_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class mni3dglmInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    file_template = traits.String("",desc="file template", usedefault=True)
    mask_template = traits.String("",desc="mask template", usedefault=True)
    design_file = File(desc='Design File')
    contrast_file = File(desc='Contrast File')
    ftest_file = File(desc='Ftest file')

class mni3dglmOutputSpec(TraitedSpec):
    group_mask = File(desc='Group Mask')
    stats_image = File(desc='4D stats image')
    stats_output_dir = traits.String(desc='stats_output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class mni3dglm_pan(BaseInterface):
    input_spec = mni3dglmInputSpec
    output_spec = mni3dglmOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = mni3dglm_proc(
            self.inputs.labels_dict,
            self.inputs.file_template,
            self.inputs.mask_template,
            self.inputs.design_file,
            self.inputs.contrast_file,
            self.inputs.ftest_file
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="mni3dglm_node",file_template="",mask_template="",design_file="",contrast_file="",ftest_file="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(mni3dglm_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
    
    # Specify node inputs
    panpipe_labels = process_fsl_glm(labels_dict)
    pan_node.inputs.labels_dict = labels_dict

    if file_template is None or file_template == "":
        file_template = getParams(labels_dict,"STAT_FILE_TEMPLATE")

    if mask_template is None or mask_template == "":
        mask_template  = getParams(labels_dict,"STAT_MASK_TEMPLATE")

    if design_file is None or design_file == "" or not os.path.isfile(design_file):
        design_file  = getParams(labels_dict,"FSL_DESIGN")

    if contrast_file is None or contrast_file == "" or not os.path.isfile(contrast_file):
        contrast_file = getParams(labels_dict,"FSL_CONTRAST")

    if ftest_file is None or ftest_file == "" or not os.path.isfile(ftest_file):
        ftest_file = getParams(labels_dict,"FSL_FTEST")


    pan_node.inputs.file_template =  file_template
    pan_node.inputs.mask_template =  mask_template
    pan_node.inputs.design_file =  design_file
    pan_node.inputs.contrast_file =  contrast_file
    pan_node.inputs.ftest_file =  ftest_file

    return pan_node


